#  Copyright (C) 2021, Christoph Peters, Karlsruhe Institute of Technology
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


from glsl_math import *

def solve_cubic(out_roots, coeffs):
    coeffs[0:3] /= coeffs[3]
    coeffs[1:3] /= 3.0
    delta = vec3(
        fma(-coeffs[2], coeffs[2], coeffs[1]),
        fma(-coeffs[1], coeffs[2], coeffs[0]),
        coeffs[2] * coeffs[0] - coeffs[1] * coeffs[1])
    discriminant = 4.0 * delta[0] * delta[2] - delta[1] * delta[1]
    sqrt_abs_discriminant = sqrt(abs(discriminant))
    depressed = vec2(fma(-2.0 * coeffs[2], delta[0], delta[1]), delta[0])
    if (discriminant >= 0.0):
        theta = arctan2(sqrt_abs_discriminant, -depressed[0]) * (1.0 / 3.0)
        cubic_root = vec2(cos(theta), sin(theta))
        out_roots[0:3] = vec3(
            cubic_root[0],
            fma(-sqrt(0.75), cubic_root[1], -0.5 * cubic_root[0]),
            fma(+sqrt(0.75), cubic_root[1], -0.5 * cubic_root[0]))
        out_roots[0:3] = fma(vec3(2.0 * sqrt(-depressed[1])), out_roots, -vec3(coeffs[2]))
        return True
    else:
        signed_sqrt_discriminant = where((depressed[0] < 0.0) ,sqrt_abs_discriminant ,-sqrt_abs_discriminant)
        quadratic_root = 0.5 * (signed_sqrt_discriminant - depressed[0])
        cube_roots = zeros((2,))
        cube_roots[0] = pow(abs(quadratic_root), 1.0 / 3.0)
        cube_roots[0] = where((quadratic_root < 0.0) ,-cube_roots[0] ,cube_roots[0])
        cube_roots[1] = -depressed[1] / cube_roots[0]
        cubic_root = cube_roots[0] + cube_roots[1]
        out_roots[0] = cubic_root - coeffs[2]
        return False
